<?php

/**
 * @file
 *   The Menu Wizard administration pages.
 */

/**
 * List all available menu items.
 */
function mew_form_list(&$form_state) {
  drupal_add_css(drupal_get_path('module', 'mew') . '/css/mew.css');
  drupal_add_js(drupal_get_path('module', 'mew') . '/js/mew.js');

  $form['keywords'] = array(
    '#type' => 'textfield',
    '#title' => t('Filter menu items on their URL path or title'),
    '#default_value' => isset($form_state['values']) ? $form_state['values']['keywords'] : NULL,
  );
  $form['buttons']['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Filter'),
    '#attributes' => array(
      'class' => 'js-hide',
    ),
  );
  $keywords = array();
  if (isset($form_state['values']['keywords']) && strlen($form_state['values']['keywords'])) {
    $keywords = explode(' ', $form_state['values']['keywords']);
  }
  $form['list'] = array(
    '#value' => mew_list($keywords),
    '#prefix' => '<div id ="mew-list-wrapper">',
    '#suffix' => '</div>',
  );

  return $form;
}

/**
 * Form submit handler for mew_form_list().
 */
function mew_form_list_submit(array $form, array &$form_state) {
  $form_state['rebuild'] = TRUE;
}

/**
 * List all available menu items.
 */
function mew_list(array $keywords = array()) {
  $items = mew_menu_items();
  $paths_mew = array_keys(mew_menu());
  $paths_altered= array();
  $result = db_query("SELECT path FROM {mew}");
  while ($path = db_result($result)) {
    $paths_altered[] = $path;
  }

  // Sort the items.
  ksort($items);
  if (isset($_GET['order']) && $_GET['order'] == t('Link title')) {
    uasort($items, 'mew_sort_title');
  }
  if (isset($_GET['sort']) && $_GET['sort'] == 'desc') {
    $items = array_reverse($items);
  }

  $item_types = mew_list_types();
  $rows = array();
  foreach ($items as $path => $item) {
    if ($keywords) {
      $strings = array($path, $item['title']);
      $matches = array(TRUE, TRUE);
      foreach ($strings as $i => $string) {
        foreach ($keywords as $keyword) {
          if (stripos($string, $keyword) === FALSE) {
            $matches[$i] = FALSE;
          }
        }
      }
      // Not all keywords are present, so skip this menu item.
      if (!$matches[0] && !$matches[1]) {
        continue;
      }
    }

    $row = array(
      'data' => array($path, $item['title']),
    );
    // Menu Wizard's menu items cannot be edited.
    if (!in_array($path, $paths_mew)) {
      $row['data'][] = l(t('edit'), 'admin/build/mew/edit/' . $path);
      // Show a reset link for altered items or an empty cell otherwise.
      if (in_array($path, $paths_altered)) {
        $row['class'] = 'mew-altered';
        $row['data'][] = l(t('reset'), 'admin/build/mew/reset/' . $path);
      }
      else {
        $row['data'][] = NULL;
      }
    }
    // Show empty cells for Menu Wizard's menu items.
    else {
      $row['data'][] = array(
        'data' => NULL,
        'colspan' => 2,
      );
    }
    $rows[] = $row;
  }

  $header = array(
    array(
      'data' => t('URL path'),
      'field' => t('path'),
    ),
    array(
      'data' => t('Title'),
      'field' => t('title'),
    ),
    array(
      'data' => t('Operations'),
      'colspan' => 2,
    ),
  );

  return theme('table', $header, $rows, array('id' => 'mew-list'));
}

/**
 * Sort menu items by title. usort() callback.
 */
function mew_sort_title(array $item_a, array $item_b) {
  return strcmp($item_a['title'], $item_b['title']);
}

/**
 * Return a list of menu item types.
 *
 * @return array
 *   Keys are menu constants and values are human-readable type titles.
 */
function mew_list_types() {
  return array(
    MENU_VISIBLE_IN_TREE => t('Visible in menu tree (MENU_VISIBLE_IN_TREE)'),
    MENU_LINKS_TO_PARENT => t('Links to same page as parent item (MENU_LINKS_TO_PARENT)'),
    MENU_IS_LOCAL_TASK => t('Tab (MENU_IS_LOCAL_TASK)'),
    MENU_VISIBLE_IN_BREADCRUMB => t('Visible in breadcrumb (MENU_CALLBACK/MENU_VISIBLE_IN_BREADCRUMB)'),
    MENU_SUGGESTED_ITEM => t("Appears at its menu's administration page (MENU_SUGGESTED_ITEM)"),
  );
}

/**
 * Return a menu item's type as an array of menu constants.
 *
 * @param $item array
 *   The menu item.
 *
 * @return array
 *   Items are one or more constants from mew_list_types().
 */
function mew_item_type($item) {
  $types = array();
  foreach (array_keys(mew_list_types()) as $type) {
    if (($item['type'] & $type) == $type) {
      $types[] = $type;
    }
  }

  return $types;
}

/**
 * Menu link edit form.
 */
function mew_form_edit(&$form_state) {
  // The first 21 characters are "admin/build/mew/edit/", which is the menu
  // item for this form.
  $path = substr($_GET['q'], 21);
  if (!($item = mew_menu_item($path))) {
    drupal_not_found();
    exit;
  }

  drupal_set_title(t('Edit %url_path', array('%url_path' => $path)));
  $form_state['storage']['item'] = $item;

  $form['path'] = array(
    '#type' => 'value',
    '#value' => $path,
  );
  $form['type'] = array(
    '#type' => 'checkboxes',
    '#title' => t('Link type'),
    '#options' => mew_list_types(),
    '#default_value' => mew_item_type($item),
    '#required' => TRUE,
  );
  $form['file'] = array(
    '#type' => 'textfield',
    '#title' => t('Include file'),
    '#default_value' => $item['file'],
  );
  $callbacks = array(
    'title' => array(
      '#title' => t('Title callback'),
      '#callback_description' => t('Defaults to <code>!t</code>. Enter <code>FALSE</code> to keep the title untranslated.', array('!t' => l('t()', 'http://api.drupal.org/api/search/6/t'))),
      '#arguments_description' => t("One argument per line. Integers will match the corresponding URL path component. The menu item's title will automatically be the first argument."),
    ),
    'page' => array(
      '#title' => t('Page callback'),
      '#callback_description' => NULL,
      '#arguments_description' => t('One argument per line. Integers will match the corresponding URL path component.'),
    ),
    'access' => array(
      '#title' => t('Access callback'),
      '#callback_description' => t('Enter <code>TRUE</code> to allow access to everyone.'),
      '#arguments_description' => t('One argument per line. Integers will match the corresponding URL path component.'),
    ),
  );
  $booleans = array('FALSE', 'TRUE');
  foreach ($callbacks as $callback => $callback_info) {
    $form[$callback] = array(
      '#type' => 'fieldset',
      '#title' => $callback_info['#title'],
      '#collapsible' => TRUE,
      '#collapsed' => TRUE,
    );

    // The callback only has primitive arguments.
    if (!mew_callback_nonprimitive($item, $callback)) {
      $form[$callback][$callback . '_callback'] = array(
        '#type' => 'textfield',
        '#title' => t('Function name'),
        '#description' => $callback_info['#callback_description'],
        '#default_value' => is_bool($item[$callback . '_callback']) ? $booleans[(bool) $item[$callback . '_callback']] : $item[$callback . '_callback'],
      );
      // Don't show the arguments element for the delivery callback, which
      // has one fixed argument.
      if ($callback != 'delivery') {
        $disabled = FALSE;
        foreach ($item[$callback . '_arguments'] as $i => $argument) {
          if (is_object($argument) || is_array($argument)) {
            $item[$callback . '_arguments'][$i] = '#' . gettype($argument) . '#';
            $disabled = TRUE;
          }
        }
        $form[$callback][$callback . '_arguments'] = array(
          '#type' => 'textarea',
          '#title' => t('Function arguments'),
          '#description' => $callback_info['#arguments_description'],
          '#default_value' => implode("\n", $item[$callback . '_arguments']),
          '#disabled' => $disabled,
        );
      }
    }
    // The callback has at least one non-primitive argument.
    else {
      $form[$callback]['#description'] = t('This callback cannot be altered because it uses object or array arguments.');
    }
  }
  $form['buttons']['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Save'),
  );
  if (db_result(db_query("SELECT COUNT(*) FROM {mew} WHERE path = '%s'", $path))) {
    $form['buttons']['reset'] = array(
      '#type' => 'markup',
      '#value' => l(t('Reset'), 'admin/build/mew/reset/' . $path),
    );
  }

  return $form;
}

/**
 * Check if a callback has non-primitive arguments.
 *
 * @param $item array
 *   The menu item to check.
 * @param $callback string
 *   The callback name.
 *
 * @return boolean
 */
function mew_callback_nonprimitive(array $item, $callback) {
  // The delivery callback accepts no customizable arguments at all, so there
  // can be no non-primitive ones.
  if ($callback == 'delivery') {
    return FALSE;
  }
  foreach ($item[$callback . '_arguments'] as $argument) {
    if (is_object($argument) || is_array($argument)) {
      return TRUE;
    }
  }
  return FALSE;
}


/**
 * Submit callback for mew_form_edit().
 */
function mew_form_edit_validate(array $form, array &$form_state) {
  $values = $form_state['values'];

  // Confirm that the include file exists.
  if (!empty($values['file'])) {
    if (!file_exists($values['file'])) {
      form_set_error('file', t('The include file does not exist.'));
    }
    else{
      require_once($values['file']);
      _mew_form_edit_validate_callbacks($form_state);
    }
  }
  else {
    _mew_form_edit_validate_callbacks($form_state);
  }
}

/**
 * Validate the callback functions entered in mew_form_edit().
 *
 * @param $form_state array
 *   $form_state as passed on to mew_form_edit_validate().
 *
 * @return NULL
 */
function _mew_form_edit_validate_callbacks(array $form_state) {
  $values = $form_state['values'];
  $error_name = array(
    'title' => t('The title callback cannot be found.'),
    'page' => t('The page callback cannot be found.'),
    'access' => t('The access callback cannot be found.'),
  );
  $name_fixed_values = array(
    'title' => array('FALSE'),
    'access' => array('TRUE'),
  );
  $error_arguments = array(
    'title' => 'The title callback requires at least !count argument(s).',
    'page' => 'The page callback requires at least !count argument(s).',
    'access' => 'The access callback requires at least !count argument(s).',
  );
  foreach (array_keys($error_name) as $callback) {
    // Check this callback has no non-primitive arguments.
    if (!mew_callback_nonprimitive($form_state['storage']['item'], $callback)) {
      $callback_name = $values[$callback . '_callback'];
      // Only validate non-empty function names.
      if (!empty($callback_name)) {
        // Only validate if the function name is no fixed value.
        if (!isset($name_fixed_values[$callback]) || !in_array($callback_name, $name_fixed_values[$callback])) {
          // Validate the callback function exists.
          if (!function_exists($callback_name)) {
            form_set_error($callback . '_callback', $error_name[$callback]);
          }
          // If the callback function exists, validate the arguments.
          elseif (isset($error_arguments[$callback])) {
            $reflection = new ReflectionFunction($callback_name);
            $count = $reflection->getNumberOfRequiredParameters();
            // The title callback has one required parameter.
            if ($callback == 'title') {
              $count--;
            }
            if (count(preg_split('#\v#', $values[$callback . '_arguments'], NULL, PREG_SPLIT_NO_EMPTY)) < $count) {
              form_set_error($callback . '_arguments', t($error_arguments[$callback], array('!count' => $count)));
            }
          }
        }
      }
    }
  }
}

/**
 * Submit callback for mew_form_edit().
 */
function mew_form_edit_submit(array $form, array &$form_state) {
  $values = $form_state['values'];
  $item = $form_state['storage']['item'];
  unset($form_state['storage']);

  // Calculate the item's type.
  $type = 0;
  foreach (array_keys(array_filter($values['type'])) as $flag) {
    $type = $type | $flag;
  }
  $item['type'] = $type;

  // Set callbacks and their arguments.
  foreach (array('title', 'page', 'access') as $callback) {
    if (!mew_callback_nonprimitive($item, $callback)) {
      $item[$callback . '_callback'] = $values[$callback . '_callback'];
      $item[$callback . '_arguments'] = preg_split('#\v#', $values[$callback . '_arguments'], NULL, PREG_SPLIT_NO_EMPTY);
      // Convert numbers to integers, so they match URL path components.
      foreach ($item[$callback . '_arguments'] as &$argument) {
        if (!preg_match('#\D#', $argument)) {
          $argument = (int) $argument;
        }
      }
    }
  }

  // The title callback defaults to t() and can be FALSE to keep the title
  // untranslated.
  if (!mew_callback_nonprimitive($item, 'title')) {
    if (empty($values['title_callback'])) {
      $item['title_callback'] = 't';
    }
    elseif ($item['title_callback'] == 'FALSE') {
      $item['title_callback'] = FALSE;
    }
  }
  // The access callback can be TRUE to allow access to everyone.
  if (!mew_callback_nonprimitive($item, 'access') && $values['access_callback'] == 'TRUE') {
    $item['access_callback'] = TRUE;
  }

  mew_save($item);
  mew_menu_rebuild();
  drupal_set_message(t('The menu item for %path has been altered.', array('%path' => $item['path'])));
  $form_state['redirect'] = 'admin/build/mew';
}

/**
 * Form: reset a menu item.
 *
 * @return array
 *   A Drupal form.
 */
function mew_form_reset(array &$form_state) {
  // The first 22 characters are "admin/build/mew/reset/", which is the menu
  // item for this form.
  $path = substr($_GET['q'], 22);

  if (!mew_menu_item($path)) {
    drupal_not_found();
    exit;
  }

  $form['path'] = array(
    '#type' => 'value',
    '#value' => $path,
  );

  return confirm_form($form,
    t('Are you sure you want to reset %title?', array('%title' => $path)),
    isset($_GET['destination']) ? $_GET['destination'] : 'admin/build/mew',
    t('This action cannot be undone.'),
    t('Reset'),
    t('Cancel')
  );
}

/**
 * Submit callback for mew_form_reset().
 */
function mew_form_reset_submit(array $form, array &$form_state) {
  mew_delete($form_state['values']['path']);
  mew_menu_rebuild();
  drupal_set_message(t('The menu item for %path has been reset.', array('%path' => $form_state['values']['path'])));
  $form_state['redirect'] = 'admin/build/mew';
}

/**
 * Rebuild the menu after changing menu items.
 */
function mew_menu_rebuild() {
  // Menu items can appear twice after altering them. The following
  // workaround prevents this: disable all modules, rebuild the menu, enable
  // all modules and rebuild the menu again. This doesn't work if all modules
  // are disabled, because then there are no menu items, so we do this
  // separately for Menu Wizard.
  // Get all currently set messages so we can set them again later to hide
  // messages that are set when enabling modules.
  $messages = drupal_get_messages();
  $modules = module_list();
  unset($modules['mew']);
  module_disable($modules);
  menu_rebuild();
  module_enable($modules);
  menu_rebuild();
  module_disable(array('mew'));
  menu_rebuild();
  module_enable(array('mew'));
  menu_rebuild();
  // Clear messages that were set in the meantime and set the old ones again.
  drupal_get_messages();
  foreach ($messages as $type => $type_messages) {
    foreach ($type_messages as $message) {
      drupal_set_message($message, $type);
    }
  }
}